"""
# all tests works
"""
import sys

sys.path.append("../")
import NEMtropy.graph_classes as sample
from NEMtropy.models_functions import *
import NEMtropy.matrix_generator as mg
import NEMtropy.solver_functions as sof
import numpy as np
import unittest  # test tool


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_iterative_0(self):

        n, seed = (4, 22)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = 100
        args = g.args

        fun = lambda x: iterative_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=700,
            method="fixed-point",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=False,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 0: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    def test_iterative_1(self):

        n, seed = (4, 22)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a[0, :] = 0

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = 100
        args = g.args

        fun = lambda x: iterative_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=700,
            method="fixed-point",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=False,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 0: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    @unittest.skip("skip large graph")
    def test_iterative_2(self):

        n, seed = (40, 22)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = 1e3
        args = g.args

        fun = lambda x: iterative_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=700,
            method="fixed-point",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 1: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    @unittest.skip("skip large graph")
    def test_iterative_3(self):

        n, seed = (40, 22)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a[0, :] = 0

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = 1e3
        args = g.args

        fun = lambda x: iterative_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function
        
        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=700,
            method="fixed-point",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 2: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    def test_newton_0(self):

        n, seed = (4, 26)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = np.infty
        args = g.args

        fun = lambda x: -loglikelihood_prime_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        fun_jac = lambda x: -loglikelihood_hessian_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (loglikelihood_dcm_exp, args))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=100,
            method="newton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser=hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 3: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    def test_newton_1(self):

        n, seed = (4, 26)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a[0, :] = 0

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = np.infty
        args = g.args

        fun = lambda x: -loglikelihood_prime_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        fun_jac = lambda x: -loglikelihood_hessian_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=100,
            method="newton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 3: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    @unittest.skip("skip large graph")
    def test_newton_1(self):

        n, seed = (40, 56)
        # n, seed = (400, 26)
        # n, seed = (4, 26)
        # n, seed = (5, 26)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = np.infty
        args = g.args

        fun = lambda x: -loglikelihood_prime_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        fun_jac = lambda x: -loglikelihood_hessian_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-1,
            max_steps=100,
            method="newton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 4: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    @unittest.skip("skip large graph")
    def test_newton_2(self):

        n, seed = (40, 56)
        # n, seed = (400, 26)
        # n, seed = (4, 26)
        # n, seed = (5, 26)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a[0, :] = 0

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = np.infty
        args = g.args

        fun = lambda x: -loglikelihood_prime_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        fun_jac = lambda x: -loglikelihood_hessian_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-1,
            max_steps=100,
            method="newton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 5: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    def test_quasinewton_0(self):

        n, seed = (4, 26)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = np.infty
        args = g.args

        fun = lambda x: -loglikelihood_prime_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        fun_jac = lambda x: -loglikelihood_hessian_diag_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (loglikelihood_dcm_exp, args))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-1,
            max_steps=100,
            method="quasinewton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 6: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

        # debug
        # print(ek)
        # print(k)
        # print('\ntest 7: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    def test_quasinewton_1(self):

        n, seed = (4, 26)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a[0, :] = 0

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = np.infty
        args = g.args

        fun = lambda x: -loglikelihood_prime_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        fun_jac = lambda x: -loglikelihood_hessian_diag_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (loglikelihood_dcm_exp, args))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-1,
            max_steps=100,
            method="quasinewton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 6: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

        # debug
        # print(ek)
        # print(k)
        # print('\ntest 7: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    @unittest.skip("skip large graph")
    def test_quasinewton_2(self):

        n, seed = (40, 26)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = np.infty
        args = g.args

        fun = lambda x: -loglikelihood_prime_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        fun_jac = lambda x: -loglikelihood_hessian_diag_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-1,
            max_steps=100,
            method="quasinewton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 6: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

        # debug
        # print(ek)
        # print(k)
        # print('\ntest 7: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

    @unittest.skip("skip large graph")
    def test_quasinewton_3(self):

        n, seed = (40, 26)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a[0, :] = 0

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)
        """
        nz_index_out = np.array([0,1,2])
        nz_index_in = np.array([0,1,2,3])
        c = np.array([2,1,1])
        """
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "random"
        g.full_return = False
        g._set_initial_guess("dcm")
        g._set_args("dcm")

        x0 = g.x0
        x0[x0 == 0] = np.infty
        args = g.args

        fun = lambda x: -loglikelihood_prime_dcm_exp(x, args)
        step_fun = lambda x: -loglikelihood_dcm_exp(x, args)
        fun_jac = lambda x: -loglikelihood_hessian_diag_dcm_exp(x, args)
        lin_fun = lambda x: linsearch_fun_DCM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        f = fun(x0)
        norm = np.linalg.norm(f)

        theta_sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-1,
            max_steps=100,
            method="quasinewton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        g._set_solved_problem_dcm(theta_sol)
        theta_sol_full = np.concatenate((g.x, g.y))

        sol = np.exp(-theta_sol_full)
        sol[np.isnan(sol)] = 0

        ek = np.concatenate(
            (
                expected_out_degree_dcm(sol),
                expected_in_degree_dcm(sol),
            )
        )
        k = np.concatenate((k_out, k_in))
        err = np.max(abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 6: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)

        # debug
        # print(ek)
        # print(k)
        # print('\ntest 7: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-2)


if __name__ == "__main__":
    unittest.main()



import sys
import os
sys.path.append("../")
import NEMtropy.graph_classes as sample
import NEMtropy.matrix_generator as mg
import numpy as np
import unittest  # test tool
import random
import networkx as nx


class MyTest(unittest.TestCase):


    def setUp(self):
        pass

    @unittest.skip("tmp")
    def test_0(self):
        N, seed = (10, 42)
        network = mg.random_weighted_matrix_generator_dense(
            n=N, sup_ext=10, sym=False, seed=seed, intweights=False
        )

        g = sample.DirectedGraph(network)
        network_bin = (network > 0).astype(int)

        g.solve_tool(
            model="crema",
            method="quasinewton",
            initial_guess="random",
            adjacency=network_bin,
            max_steps=1000,
            verbose=False,
        )

        # g._solution_error()
        err = g.error

        # print('\ntest 5: error = {}'.format(g.error))
        n = 100
        output_dir = "sample_crema_decm_det/"
        # random.seed(100)
        g.ensemble_sampler(n=n, output_dir=output_dir, seed=42)

        #

        dk_out = {'{}'.format(i): g.dseq_out[i] for i in range(N)}
        dk_in = {'{}'.format(i): g.dseq_in[i] for i in range(N)}
        ds_out = {'{}'.format(i): g.out_strength[i] for i in range(N)}
        ds_in = {'{}'.format(i): g.in_strength[i] for i in range(N)}

        # read all sampled graphs and check the average degree distribution is close enough
        dk_out_emp = {'{}'.format(i): 0 for i in range(N)}
        dk_in_emp = {'{}'.format(i): 0 for i in range(N)}
        ds_out_emp = {'{}'.format(i): 0 for i in range(N)}
        ds_in_emp = {'{}'.format(i): 0 for i in range(N)}

        for l in range(n):
            f = output_dir + "{}.txt".format(l)
            if not os.stat(f).st_size == 0:
                g_tmp = nx.read_edgelist(f, data=(
                    ("weight", float),), create_using=nx.DiGraph())
                dk_out_tmp = dict(g_tmp.out_degree)
                dk_in_tmp = dict(g_tmp.in_degree)
                ds_out_tmp = dict(g_tmp.out_degree(weight='weight'))
                ds_in_tmp = dict(g_tmp.in_degree(weight='weight'))
                for item in dk_out_tmp.keys():
                    dk_out_emp[item] += dk_out_tmp[item]
                    dk_in_emp[item] += dk_in_tmp[item]
                    ds_out_emp[item] += ds_out_tmp[item]
                    ds_in_emp[item] += ds_in_tmp[item]

        for item in dk_out_emp.keys():
            dk_out_emp[item] = dk_out_emp[item]/n
            dk_in_emp[item] = dk_in_emp[item]/n
            ds_out_emp[item] = ds_out_emp[item]/n
            ds_in_emp[item] = ds_in_emp[item]/n

        adk_out_diff = np.array(
            [abs(dk_out[item] - dk_out_emp[item]) for item in dk_out.keys()])
        adk_in_diff = np.array(
            [abs(dk_in[item] - dk_in_emp[item]) for item in dk_in.keys()])
        ads_out_diff = np.array(
            [abs(ds_out[item] - ds_out_emp[item]) for item in ds_out.keys()])
        ads_in_diff = np.array(
            [abs(ds_in[item] - ds_in_emp[item]) for item in ds_in.keys()])
        a_diff = np.concatenate(
            (adk_out_diff, adk_in_diff, ads_out_diff, ads_in_diff))
        # d_diff = {item:d[item] - d_emp[item] for item in d.keys()}
        # s_diff = {item:s[item] - s_emp[item] for item in s.keys()}

        ensemble_error = np.linalg.norm(a_diff, np.inf)

        # debug
        """
        for i in range(N):
            for j in range(N):
                if i!=j:
                    aux = x[i]*x[j]
                    # print("({},{}) p = {}".format(i,j,aux/(1+aux)))
        """

        # debug
        """
        print('\n original degree sequence ', dk_out, dk_in)
        print('\n original strength sequence ',ds_out, ds_in)
        print('\n ensemble average degree sequence', dk_out_emp, dk_in_emp)
        print('\n ensemble average strength sequence', ds_out_emp, ds_in_emp)
        print('\n empirical error = {}'.format(ensemble_error))
        print('\n theoretical error = {}'.format(err))
        print('\n original degree sequence ', dk_out, dk_in)
        """

        l = os.listdir(output_dir)

        for f in l:
            os.remove(output_dir + f)
        os.rmdir(output_dir)


        # test result
        self.assertTrue(ensemble_error < 4)

    def test_1(self):
        N, seed = (5, 42)
        """
        network = mg.random_weighted_matrix_generator_dense(
            n=N, sup_ext=10, sym=False, seed=seed, intweights=False
        )
        """
        network = mg.random_weighted_matrix_generator_uniform_custom_density(
            n=N,
            p=0.2,
            sym=False,
            sup_ext=30,
            intweights=True,
            seed=seed
            )


        g = sample.DirectedGraph(network)
        network_bin = (network > 0).astype(int)

        g.solve_tool(
            model="crema",
            method="quasinewton",
            initial_guess="random",
            adjacency=network_bin,
            max_steps=1000,
            verbose=False,
        )

        # g._solution_error()
        err = g.error

        # print('\ntest 5: error = {}'.format(g.error))
        n = 100
        output_dir = "sample_crema_decm_det/"
        # random.seed(100)
        g.ensemble_sampler(n=n, output_dir=output_dir, seed=42)

        #

        dk_out = {'{}'.format(i): g.dseq_out[i] for i in range(N)}
        dk_in = {'{}'.format(i): g.dseq_in[i] for i in range(N)}
        ds_out = {'{}'.format(i): g.out_strength[i] for i in range(N)}
        ds_in = {'{}'.format(i): g.in_strength[i] for i in range(N)}

        # read all sampled graphs and check the average degree distribution is close enough
        dk_out_emp = {'{}'.format(i): 0 for i in range(N)}
        dk_in_emp = {'{}'.format(i): 0 for i in range(N)}
        ds_out_emp = {'{}'.format(i): 0 for i in range(N)}
        ds_in_emp = {'{}'.format(i): 0 for i in range(N)}

        for l in range(n):
            f = output_dir + "{}.txt".format(l)
            if not os.stat(f).st_size == 0:
                g_tmp = nx.read_edgelist(f, data=(
                    ("weight", float),), create_using=nx.DiGraph())
                dk_out_tmp = dict(g_tmp.out_degree)
                dk_in_tmp = dict(g_tmp.in_degree)
                ds_out_tmp = dict(g_tmp.out_degree(weight='weight'))
                ds_in_tmp = dict(g_tmp.in_degree(weight='weight'))
                for item in dk_out_tmp.keys():
                    dk_out_emp[item] += dk_out_tmp[item]
                    dk_in_emp[item] += dk_in_tmp[item]
                    ds_out_emp[item] += ds_out_tmp[item]
                    ds_in_emp[item] += ds_in_tmp[item]

        for item in dk_out_emp.keys():
            dk_out_emp[item] = dk_out_emp[item]/n
            dk_in_emp[item] = dk_in_emp[item]/n
            ds_out_emp[item] = ds_out_emp[item]/n
            ds_in_emp[item] = ds_in_emp[item]/n

        adk_out_diff = np.array(
            [abs(dk_out[item] - dk_out_emp[item]) for item in dk_out.keys()])
        adk_in_diff = np.array(
            [abs(dk_in[item] - dk_in_emp[item]) for item in dk_in.keys()])
        ads_out_diff = np.array(
            [abs(ds_out[item] - ds_out_emp[item]) for item in ds_out.keys()])
        ads_in_diff = np.array(
            [abs(ds_in[item] - ds_in_emp[item]) for item in ds_in.keys()])
        a_diff = np.concatenate(
            (adk_out_diff, adk_in_diff, ads_out_diff, ads_in_diff))
        # d_diff = {item:d[item] - d_emp[item] for item in d.keys()}
        # s_diff = {item:s[item] - s_emp[item] for item in s.keys()}

        ensemble_error = np.linalg.norm(a_diff, np.inf)

        # debug
        """
        for i in range(N):
            for j in range(N):
                if i!=j:
                    aux = x[i]*x[j]
                    # print("({},{}) p = {}".format(i,j,aux/(1+aux)))
        """

        # debug
        """
        print('\n original degree sequence ', dk_out, dk_in)
        print('\n original strength sequence ',ds_out, ds_in)
        print('\n ensemble average degree sequence', dk_out_emp, dk_in_emp)
        print('\n ensemble average strength sequence', ds_out_emp, ds_in_emp)
        print('\n empirical error = {}'.format(ensemble_error))
        print('\n theoretical error = {}'.format(err))
        print('\n original degree sequence ', dk_out, dk_in)
        """

        l = os.listdir(output_dir)

        for f in l:
            os.remove(output_dir + f)
        os.rmdir(output_dir)


        # test result
        self.assertTrue(ensemble_error < 4)



if __name__ == "__main__":
    unittest.main()



import sys
import os
sys.path.append("../")
import NEMtropy.graph_classes as sample
import NEMtropy.matrix_generator as mg
import numpy as np
import unittest  # test tool
import random
import networkx as nx


class MyTest(unittest.TestCase):

    def setUp(self):
        pass

    @unittest.skip("tmp")
    def test_0(self):
        n, seed = (5, 42)
        network = mg.random_weighted_matrix_generator_dense(
            n=n, sup_ext=10, sym=True, seed=seed, intweights=True
        )
        # number of copies to generate

        g = sample.UndirectedGraph(adjacency=network)

        g.solve_tool(
            model="crema-sparse",
            method="quasinewton",
            initial_guess="random",
            adjacency="cm",
            max_steps=1000,
            verbose=False,
        )

        # g._solution_error()
        err = g.error

        # print('\ntest 5: error = {}'.format(g.error))
        n_sample = 50
        output_dir = "sample_crema_ecm_prob/"
        # random.seed(100)
        g.ensemble_sampler(n=n_sample, output_dir=output_dir, seed=42)


        d = {'{}'.format(i): g.dseq[i] for i in range(n)}
        s = {'{}'.format(i): g.strength_sequence[i] for i in range(n)}

        # read all sampled graphs and check the average degree distribution
        d_emp = {'{}'.format(i): 0 for i in range(n)}
        s_emp = {'{}'.format(i): 0 for i in range(n)}

        for l in range(n_sample):
            f = output_dir + "{}.txt".format(l)
            if not os.stat(f).st_size == 0:
                g_tmp = nx.read_edgelist(f, data=(("weight", float),))
                d_tmp = dict(g_tmp.degree)
                s_tmp = dict(g_tmp.degree(weight='weight'))
                for item in d_tmp.keys():
                    d_emp[item] += d_tmp[item]
                    s_emp[item] += s_tmp[item]

        for item in d_emp.keys():
            d_emp[item] = d_emp[item]/n_sample
            s_emp[item] = s_emp[item]/n_sample

        ad_diff = np.array([abs(d[item] - d_emp[item]) for item in d.keys()])
        as_diff = np.array([abs(s[item] - s_emp[item]) for item in s.keys()])
        a_diff = np.concatenate((ad_diff, as_diff))
        d_diff = {item: abs(d[item] - d_emp[item]) for item in d.keys()}
        s_diff = {item: abs(s[item] - s_emp[item]) for item in s.keys()}

        ensemble_error = np.linalg.norm(a_diff, np.inf)

        # debug
        # print('\n original degree sequence ', d)
        # print('\n original strength sequence ', s)
        # print('\n ensemble average strength sequence', s_emp)
        # print('\n degree by degree difference vector ', d_diff)
        # print('\n strength by strength difference vector ', s_diff)
        # print('\n empirical error = {}'.format(ensemble_error))
        # print('\n theoretical error = {}'.format(err))

        l = os.listdir(output_dir)
        for f in l:
            os.remove(output_dir + f)
        os.rmdir(output_dir)

        # test result
        self.assertTrue(ensemble_error < 3)

    def test_1(self):
        n, seed = (10, 42)
        """
        network = mg.random_weighted_matrix_generator_dense(
            n=n, sup_ext=10, sym=True, seed=seed, intweights=True
        )
        """
        network = mg.random_weighted_matrix_generator_uniform_custom_density(
            n=n,
            p=0.2,
            sym=True,
            sup_ext=10,
            intweights=True,
            seed=seed
        )

        # number of copies to generate

        g = sample.UndirectedGraph(adjacency=network)

        g.solve_tool(
            model="crema-sparse",
            method="quasinewton",
            initial_guess="random",
            adjacency="cm",
            max_steps=1000,
            verbose=False,
        )

        # g._solution_error()
        err = g.error

        # print('\ntest 5: error = {}'.format(g.error))
        n_sample = 500
        output_dir = "sample_crema_ecm_prob/"
        # random.seed(100)
        g.ensemble_sampler(n=n_sample, output_dir=output_dir, seed=42)


        d = {'{}'.format(i): g.dseq[i] for i in range(n)}
        s = {'{}'.format(i): g.strength_sequence[i] for i in range(n)}

        # read all sampled graphs and check the average degree distribution
        d_emp = {'{}'.format(i): 0 for i in range(n)}
        s_emp = {'{}'.format(i): 0 for i in range(n)}

        for l in range(n_sample):
            f = output_dir + "{}.txt".format(l)
            if not os.stat(f).st_size == 0:
                g_tmp = nx.read_edgelist(f, data=(("weight", float),))
                d_tmp = dict(g_tmp.degree)
                s_tmp = dict(g_tmp.degree(weight='weight'))
                for item in d_tmp.keys():
                    d_emp[item] += d_tmp[item]
                    s_emp[item] += s_tmp[item]

        for item in d_emp.keys():
            d_emp[item] = d_emp[item]/n_sample
            s_emp[item] = s_emp[item]/n_sample

        ad_diff = np.array([abs(d[item] - d_emp[item]) for item in d.keys()])
        as_diff = np.array([abs(s[item] - s_emp[item]) for item in s.keys()])
        a_diff = np.concatenate((ad_diff, as_diff))
        d_diff = {item: abs(d[item] - d_emp[item]) for item in d.keys()}
        s_diff = {item: abs(s[item] - s_emp[item]) for item in s.keys()}

        ensemble_error = np.linalg.norm(a_diff, np.inf)

        # debug
        # print('\n original degree sequence ', d)
        # print('\n original strength sequence ', s)
        # print('\n ensemble average strength sequence', s_emp)
        # print('\n degree by degree difference vector ', d_diff)
        # print('\n strength by strength difference vector ', s_diff)
        # print('\n empirical error = {}'.format(ensemble_error))
        # print('\n theoretical error = {}'.format(err))

        l = os.listdir(output_dir)
        for f in l:
            os.remove(output_dir + f)
        os.rmdir(output_dir)

        # test result
        self.assertTrue(ensemble_error < 13)



if __name__ == "__main__":
    unittest.main()



import bicm
from .graph_classes import *
from .matrix_generator import *
from .ensemble_generator import *
from .models_functions import *
from .network_functions import *
from bicm.graph_classes import *
from bicm.models_functions import *
from bicm.network_functions import *

__version__ = "3.0.1"
__author__ = """Nicolo' Vallarano (nico.vallarano@gmail.com)
Emiliano Marchese (emilianomarcheserc@gmail.com)
Matteo Bruno (matteobruno180@gmail.com)"""



import numpy as np
import scipy
import scipy.sparse
from numba import jit
from scipy.sparse import csr_matrix
from bicm.network_functions import *


def build_adjacency_from_edgelist(
        edgelist,
        is_directed,
        is_sparse=False,
        is_weighted=False):
    """Generates adjacency matrix given edgelist.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_directed: True if edge direction is informative.
    :type is_directed: bool
    :param is_sparse: If true the output is a sparse matrix.
    :type is_sparse: bool
    :param is_weighted: True if the edgelist is weighted
    :type is_weighted: bool
    :return: Adjacency matrix.
    :rtype: numpy.ndarray or scipy.sparse.csr_matrix
    """
    if not isinstance(edgelist, np.ndarray):
        raise TypeError("edgelist must be an numpy.ndarray.")
    if is_sparse:
        if is_weighted:
            adjacency = build_adjacency_sparse_weighted(edgelist, is_directed)
        else:
            adjacency = build_adjacency_sparse(edgelist, is_directed)
    else:
        if is_weighted:
            adjacency = build_adjacency_fast_weighted(edgelist, is_directed)
        else:
            adjacency = build_adjacency_fast(edgelist, is_directed)
    return adjacency


@jit(nopython=True)
def build_adjacency_fast(edgelist, is_directed):
    """Generates adjacency matrix given edgelist, numpy array format
    is used.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_directed: True if edge direction is informative.
    :type is_directed: bool
    :return: Adjacency matrix.
    :rtype: numpy.ndarray
    """
    if is_directed:
        n_nodes = len(set(edgelist[:, 0]) | set(edgelist[:, 1]))
        adj = np.zeros((n_nodes, n_nodes))
        for ii in np.arange(edgelist.shape[0]):
            edges = edgelist[ii]
            i = int(edges[0])
            j = int(edges[1])
            adj[i, j] = 1
    else:
        n_nodes = len(set(edgelist[:, 0]) | set(edgelist[:, 1]))
        adj = np.zeros((n_nodes, n_nodes))
        for ii in np.arange(edgelist.shape[0]):
            edges = edgelist[ii]
            i = int(edges[0])
            j = int(edges[1])
            adj[i, j] = 1
            adj[j, i] = 1
    return adj


def build_adjacency_sparse(edgelist, is_directed):
    """Generates adjacency matrix given edgelist, scipy sparse format
    is used.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_sparse: True if edge direction is informative.
    :type is_sparse: bool
    :return: Adjacency matrix.
    :rtype: scipy.sparse.csr_matrix
    """
    n_nodes = set(edgelist[:, 0]) | set(edgelist[:, 1])
    if is_directed:
        row = edgelist[:, 0].astype(int)
        columns = edgelist[:, 1].astype(int)
        data = np.ones(n_nodes, dtype=int)
        adj = csr_matrix((data, (row, columns)), shape=(n_nodes, n_nodes))
    else:
        row = np.concatenate([edgelist[:, 0], edgelist[:, 1]]).astype(int)
        columns = np.concatenate([edgelist[:, 1], edgelist[:, 0]]).astype(int)
        data = np.ones(2 * n_nodes, dtype=int)
        adj = csr_matrix((data, (row, columns)), shape=(n_nodes, n_nodes))
    return adj


@jit(nopython=True)
def build_adjacency_fast_weighted(edgelist, is_directed):
    """Generates weighted adjacency matrix given edgelist,
    numpy array format is used.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_directed: True if edge direction is informative.
    :type is_directed: bool
    :return: Adjacency matrix.
    :rtype: numpy.ndarray
    """
    if is_directed:
        n_nodes = len(set(edgelist[:, 0]) | set(edgelist[:, 1]))
        adj = np.zeros((n_nodes, n_nodes))
        for ii in np.arange(edgelist.shape[0]):
            edges = edgelist[ii]
            i = int(edges[0])
            j = int(edges[1])
            w = edges[2]
            adj[i, j] = w
    else:
        n_nodes = len(set(edgelist[:, 0]) | set(edgelist[:, 1]))
        adj = np.zeros((n_nodes, n_nodes))
        for ii in np.arange(edgelist.shape[0]):
            edges = edgelist[ii]
            i = int(edges[0])
            j = int(edges[1])
            w = edges[2]
            adj[i, j] = w
            adj[j, i] = w
    return adj


def build_adjacency_sparse_weighted(edgelist, is_directed):
    """Generates weighted adjacency matrix given edgelist,
    scipy sparse format is used.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_sparse: True if edge direction is informative.
    :type is_sparse: bool
    :return: Adjacency matrix.
    :rtype: scipy.sparse.csr_matrix
    """
    n_nodes = set(edgelist[:, 0]) | set(edgelist[:, 1])
    if is_directed:
        row = edgelist[:, 0].astype(int)
        columns = edgelist[:, 1].astype(int)
        data = edgelist[:, 2]
        adj = csr_matrix((data, (row, columns)), shape=(n_nodes, n_nodes))
    else:
        row = np.concatenate([edgelist[:, 0], edgelist[:, 1]]).astype(int)
        columns = np.concatenate([edgelist[:, 1], edgelist[:, 0]]).astype(int)
        data = np.concatenate([edgelist[:, 2], edgelist[:, 2]])
        adj = csr_matrix((data, (row, columns)), shape=(n_nodes, n_nodes))
    return adj


def out_degree(a):
    """Returns matrix *a* out degrees sequence.

    :param a: Adjacency matrix
    :type a: numpy.ndarray, scipy.sparse.csr.csr_matrix,
        scipy.sparse.coo.coo_matrix
    :return: Out degree sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a > 0, 1)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a > 0, 1).A1  # noqa


def in_degree(a):
    """Returns matrix *a* in degrees sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse.csr.csr_matrix,
        scipy.sparse.coo.coo_matrix
    :return: In degree sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a > 0, 0)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a > 0, 0).A1


def out_strength(a):
    """Returns matrix *a* out strengths sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse.csr.csr_matrix,
        scipy.sparse.coo.coo_matrix
    :return: Out strengths sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a, 1)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a, 1).A1


def in_strength(a):
    """Returns matrix *a* in strengths sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse.csr.csr_matrix,
        scipy.sparse.coo.coo_matrix
    :return: In strengths sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a, 0)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a, 0).A1


def degree(a):
    """Returns matrix *a* degrees sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse
    :return: Degree sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a > 0, 1)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a > 0, 1).A1


def strength(a):
    """Returns matrix *a* strengths sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse
    :return: Strengths sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a, 1)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a, 1).A1


def edgelist_from_edgelist_undirected(edgelist):
    """Creates a new edgelist with the indexes of the nodes instead of the
     names. Returns also a dictionary that keep track of the nodes and,
      depending on the type of graph, degree and strengths sequences.

    :param edgelist: edgelist.
    :type edgelist: numpy.ndarray or list
    :return: edgelist, degrees sequence, strengths sequence and
     new labels to old labels dictionary.
    :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray, dict)
    """
    edgelist = [tuple(item) for item in edgelist]
    if len(edgelist[0]) == 2:
        # nodetype = type(edgelist[0][0])
        edgelist = np.array(
            edgelist,
            dtype=np.dtype([("source", object), ("target", object)]),
        )
    else:
        # nodetype = type(edgelist[0][0])
        # weigthtype = type(edgelist[0][2])
        # Vorrei mettere una condizione sul weighttype che deve essere numerico
        edgelist = np.array(
            edgelist,
            dtype=np.dtype(
                [
                    ("source", object),
                    ("target", object),
                    ("weigth", object),
                ]
            ),
        )
    # If there is a loop we count it twice in the degree of the node.
    unique_nodes, degree_seq = np.unique(
        np.concatenate((edgelist["source"], edgelist["target"])),
        return_counts=True,
    )
    nodes_dict = dict(enumerate(unique_nodes))
    inv_nodes_dict = {v: k for k, v in nodes_dict.items()}
    if len(edgelist[0]) == 2:
        edgelist_new = [
            (inv_nodes_dict[edge[0]], inv_nodes_dict[edge[1]])
            for edge in edgelist
        ]
        edgelist_new = np.array(
            edgelist_new, dtype=np.dtype([("source", int), ("target", int)])
        )
    else:
        edgelist_new = [
            (inv_nodes_dict[edge[0]], inv_nodes_dict[edge[1]], edge[2])
            for edge in edgelist
        ]
        edgelist_new = np.array(
            edgelist_new,
            dtype=np.dtype(
                [("source", int), ("target", int), ("weight", np.float64)]
            ),
        )
    if len(edgelist[0]) == 3:
        aux_edgelist = np.concatenate(
            (edgelist_new["source"], edgelist_new["target"])
        )
        aux_weights = np.concatenate(
            (edgelist_new["weight"], edgelist_new["weight"])
        )
        strength_seq = np.array(
            [aux_weights[aux_edgelist == i].sum() for i in nodes_dict]
        )
        return edgelist_new, degree_seq, strength_seq, nodes_dict
    return edgelist_new, degree_seq, nodes_dict


def edgelist_from_edgelist_directed(edgelist):
    """Creates a new edgelists replacing nodes labels with indexes.
    Returns also two dictionaries that keep track of the
    nodes index-label relation.
    Works also on weighted graphs.

    :param edgelist: List of edges.
    :type edgelist: list
    :return: Re-indexed list of edges, out-degrees, in-degrees,
        index to label dictionary
    :rtype: (dict, numpy.ndarray, numpy.ndarray, dict)
    """
    # TODO: inserire esempio edgelist pesata edgelist binaria
    # nel docstring
    # edgelist = list(zip(*edgelist))
    edgelist = [tuple(item) for item in edgelist]
    if len(edgelist[0]) == 2:
        # nodetype = type(edgelist[0][0])
        edgelist = np.array(
            edgelist,
            dtype=np.dtype(
                [
                    ("source", object),
                    ("target", object)
                ]
            ),
        )
    else:
        # nodetype = type(edgelist[0][0])
        # weigthtype = type(edgelist[0][2])
        # Vorrei mettere una condizione sul weighttype che deve essere numerico
        edgelist = np.array(
            edgelist,
            dtype=np.dtype(
                [
                    ("source", object),
                    ("target", object),
                    ("weigth", object),
                ]
            ),
        )
    # If there is a loop we count it twice in the degree of the node.
    unique_nodes = np.unique(
        np.concatenate((edgelist["source"], edgelist["target"])),
        return_counts=False,
    )
    out_degree = np.zeros_like(unique_nodes)
    in_degree = np.zeros_like(unique_nodes)
    nodes_dict = dict(enumerate(unique_nodes))
    inv_nodes_dict = {v: k for k, v in nodes_dict.items()}
    if len(edgelist[0]) == 2:
        edgelist_new = [
            (inv_nodes_dict[edge[0]], inv_nodes_dict[edge[1]])
            for edge in edgelist
        ]
        edgelist_new = np.array(
            edgelist_new, dtype=np.dtype([("source", int), ("target", int)])
        )
    else:
        edgelist_new = [
            (inv_nodes_dict[edge[0]], inv_nodes_dict[edge[1]], edge[2])
            for edge in edgelist
        ]
        edgelist_new = np.array(
            edgelist_new,
            dtype=np.dtype(
                [("source", int), ("target", int), ("weigth", np.float64)]
            ),
        )
    out_indices, out_counts = np.unique(
        edgelist_new["source"], return_counts=True
    )
    in_indices, in_counts = np.unique(
        edgelist_new["target"], return_counts=True
    )
    out_degree[out_indices] = out_counts
    in_degree[in_indices] = in_counts
    if len(edgelist[0]) == 3:
        out_strength = np.zeros_like(unique_nodes, dtype=np.float64)
        in_strength = np.zeros_like(unique_nodes, dtype=np.float64)
        out_counts_strength = np.array(
            [
                edgelist_new[edgelist_new["source"] == i]["weigth"].sum()
                for i in out_indices
            ]
        )
        in_counts_strength = np.array(
            [
                edgelist_new[edgelist_new["target"] == i]["weigth"].sum()
                for i in in_indices
            ]
        )
        out_strength[out_indices] = out_counts_strength
        in_strength[in_indices] = in_counts_strength
        return (
            edgelist_new,
            out_degree,
            in_degree,
            out_strength,
            in_strength,
            nodes_dict,
        )
    return edgelist_new, out_degree, in_degree, nodes_dict


def count_2motif_2(a):
    """Counts number of dyads.
    :param a np.ndarray: adjacency matrix
    :return: dyads count
    :rtype: int
    """

    tmp = a + a.transpose()
    return int(len(tmp[tmp == 2]))


def count_2motif_1(a):
    """Counts number of singles.
    :param a np.ndarray: adjacency matrix
    :return: singles count
    :rtype: int
    """

    tmp = a + a.transpose()
    return int(len(tmp[tmp == 1])/2)


def count_2motif_0(a):
    """Counts number of zeros.
    :param a np.ndarray: adjacency matrix
    :return: zeros count
    :rtype: int
    """

    n = a.shape[0]
    tmp = a + a.transpose()
    if isinstance(a, np.ndarray):
        return int((n*(n-1) - np.count_nonzero(tmp)))
    if isinstance(
        a,
        (
            scipy.sparse.csr_array,
            # scipy.sparse.coo_array,
            # scipy.sparse.bsr_array,
        )
    ):
        return int((n*(n-1) - tmp.count_nonzero()))

# 3-nodes motifs


@jit(nopython=True)
def count_3motif_1_ndarray(a):
    """Counts abundance of 3-nodes motif 1.
    :param a np.ndarray: adjacency matrix
    :return: motif 1 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*a[j, i]*(1 - a[i, k])*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s
 

@jit(nopython=True)
def count_3motif_2_ndarray(a):
    """Counts abundance of 3-nodes motif 2.
    :param a np.ndarray: adjacency matrix
    :return: motif 2 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*(1 - a[j, i])*(1 - a[i, k])*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s
    

@jit(nopython=True)
def count_3motif_3_ndarray(a):
    """Counts abundance of 3-nodes motif 3.
    :param a np.ndarray: adjacency matrix
    :return: motif 3 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*(1 - a[i, k])*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s
 

@jit(nopython=True)
def count_3motif_4_ndarray(a):
    """Counts abundance of 3-nodes motif 4.
    :param a np.ndarray: adjacency matrix
    :return: motif 4 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*(1 - a[j, i])*a[i, k]*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s


@jit(nopython=True)
def count_3motif_5_ndarray(a):
    """Counts abundance of 3-nodes motif 5.
    :param a np.ndarray: adjacency matrix
    :return: motif 5 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*a[j, i]*a[i, k]*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s


@jit(nopython=True)
def count_3motif_6_ndarray(a):
    """Counts abundance of 3-nodes motif 6.
    :param a np.ndarray: adjacency matrix
    :return: motif 6 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*a[i, k]*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s


@jit(nopython=True)
def count_3motif_7_ndarray(a):
    """Counts abundance of 3-nodes motif 7.
    :param a np.ndarray: adjacency matrix
    :return: motif 7 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*(1 - a[i, k])*(1 - a[k, i])*(1 - a[j, k])*a[k, j]
    return s


@jit(nopython=True)
def count_3motif_8_ndarray(a):
    """Counts abundance of 3-nodes motif 8.
    :param a np.ndarray: adjacency matrix
    :return: motif 8 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*(1 - a[i, k])*(1 - a[k, i])*a[j, k]*a[k, j]
    return s


@jit(nopython=True)
def count_3motif_9_ndarray(a):
    """Counts abundance of 3-nodes motif 9.
    :param a np.ndarray: adjacency matrix
    :return: motif 9 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*a[j, i]*a[i, k]*(1 - a[k, i])*(1 - a[j, k])*a[k, j]
    return s


@jit(nopython=True)
def count_3motif_10_ndarray(a):
    """Counts abundance of 3-nodes motif 10.
    :param a np.ndarray: adjacency matrix
    :return: motif 10 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*a[j, i]*a[i, k]*(1 - a[k, i])*a[j, k]*a[k, j]
    return s


@jit(nopython=True)
def count_3motif_11_ndarray(a):
    """Counts abundance of 3-nodes motif 11.
    :param a np.ndarray: adjacency matrix
    :return: motif 11 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*(1 - a[j, i])*a[i, k]*(1 - a[k, i])*a[j, k]*a[k, j]
    return s


@jit(nopython=True)
def count_3motif_12_ndarray(a):
    """Counts abundance of 3-nodes motif_ 12.
    :param a np.ndarray: adjacency matrix
    :return: motif 12 count
    :rtype: int
    """

    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*a[i, k]*(1 - a[k, i])*a[j, k]*a[k, j]
    return s


@jit(nopython=True)
def count_3motif_13_ndarray(a):
    """Counts abundance of 3-nodes motif_ 13.
    :param a np.ndarray: adjacency matrix
    :return: motif 13 count
    :rtype: int
    """
    n = a.shape[0]
    n = len(a)
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*a[i, k]*a[k, i]*a[j, k]*a[k, j]
    return s



def count_3motif_1_sparse(a):
    """Counts abundance of 3-nodes motif 1.
    :param a np.ndarray: adjacency matrix
    :return: motif 1 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*a[j, i]*(1 - a[i, k])*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s
 


def count_3motif_2_sparse(a):
    """Counts abundance of 3-nodes motif 2.
    :param a np.ndarray: adjacency matrix
    :return: motif 2 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*(1 - a[j, i])*(1 - a[i, k])*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s
    


def count_3motif_3_sparse(a):
    """Counts abundance of 3-nodes motif 3.
    :param a np.ndarray: adjacency matrix
    :return: motif 3 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*(1 - a[i, k])*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s
 


def count_3motif_4_sparse(a):
    """Counts abundance of 3-nodes motif 4.
    :param a np.ndarray: adjacency matrix
    :return: motif 4 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*(1 - a[j, i])*a[i, k]*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s



def count_3motif_5_sparse(a):
    """Counts abundance of 3-nodes motif 5.
    :param a np.ndarray: adjacency matrix
    :return: motif 5 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*a[j, i]*a[i, k]*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s



def count_3motif_6_sparse(a):
    """Counts abundance of 3-nodes motif 6.
    :param a np.ndarray: adjacency matrix
    :return: motif 6 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*a[i, k]*(1 - a[k, i])*a[j, k]*(1 - a[k, j])
    return s



def count_3motif_7_sparse(a):
    """Counts abundance of 3-nodes motif 7.
    :param a np.ndarray: adjacency matrix
    :return: motif 7 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*(1 - a[i, k])*(1 - a[k, i])*(1 - a[j, k])*a[k, j]
    return s



def count_3motif_8_sparse(a):
    """Counts abundance of 3-nodes motif 8.
    :param a np.ndarray: adjacency matrix
    :return: motif 8 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*(1 - a[i, k])*(1 - a[k, i])*a[j, k]*a[k, j]
    return s



def count_3motif_9_sparse(a):
    """Counts abundance of 3-nodes motif 9.
    :param a np.ndarray: adjacency matrix
    :return: motif 9 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*a[j, i]*a[i, k]*(1 - a[k, i])*(1 - a[j, k])*a[k, j]
    return s



def count_3motif_10_sparse(a):
    """Counts abundance of 3-nodes motif 10.
    :param a np.ndarray: adjacency matrix
    :return: motif 10 count
    :rtype: int
    """

    n = a.shape[0]
    
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += (1 - a[i, j])*a[j, i]*a[i, k]*(1 - a[k, i])*a[j, k]*a[k, j]
    return s



def count_3motif_11_sparse(a):
    """Counts abundance of 3-nodes motif 11.
    :param a np.ndarray: adjacency matrix
    :return: motif 11 count
    :rtype: int
    """

    n = a.shape[0]
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*(1 - a[j, i])*a[i, k]*(1 - a[k, i])*a[j, k]*a[k, j]
    return s



def count_3motif_12_sparse(a):
    """Counts abundance of 3-nodes motif_ 12.
    :param a np.ndarray: adjacency matrix
    :return: motif 12 count
    :rtype: int
    """

    n = a.shape[0]
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*a[i, k]*(1 - a[k, i])*a[j, k]*a[k, j]
    return s



def count_3motif_13_sparse(a):
    """Counts abundance of 3-nodes motif_ 13.
    :param a np.ndarray: adjacency matrix
    :return: motif 13 count
    :rtype: int
    """
    n = a.shape[0]
    s = 0
    for i in range(n):
        for j in range(n):
            if j is not i:
                for k in range(n):
                    if k is not j and k is not i and i is not j:
                        s += a[i, j]*a[j, i]*a[i, k]*a[k, i]*a[j, k]*a[k, j]
    return s


def count_3motif_1(a):
    """Counts abundance of 3-nodes motif 1.
    :param a np.ndarray: adjacency matrix
    :return: motif 1 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_1_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_1_sparse(a)
    else:
        raise TypeError
 

def count_3motif_2(a):
    """Counts abundance of 3-nodes motif 2.
    :param a np.ndarray: adjacency matrix
    :return: motif 2 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_2_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_2_sparse(a)
    else:
        raise TypeError
    

def count_3motif_3(a):
    """Counts abundance of 3-nodes motif 3.
    :param a np.ndarray: adjacency matrix
    :return: motif 3 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_3_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_3_sparse(a)
    else:
        raise TypeError
 

def count_3motif_4(a):
    """Counts abundance of 3-nodes motif 4.
    :param a np.ndarray: adjacency matrix
    :return: motif 4 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_4_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_4_sparse(a)
    else:
        raise TypeError


def count_3motif_5(a):
    """Counts abundance of 3-nodes motif 5.
    :param a np.ndarray: adjacency matrix
    :return: motif 5 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_5_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_5_sparse(a)
    else:
        raise TypeError


def count_3motif_6(a):
    """Counts abundance of 3-nodes motif 6.
    :param a np.ndarray: adjacency matrix
    :return: motif 6 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_6_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_6_sparse(a)
    else:
        raise TypeError


def count_3motif_7(a):
    """Counts abundance of 3-nodes motif 7.
    :param a np.ndarray: adjacency matrix
    :return: motif 7 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_7_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_7_sparse(a)
    else:
        raise TypeError


def count_3motif_8(a):
    """Counts abundance of 3-nodes motif 8.
    :param a np.ndarray: adjacency matrix
    :return: motif 8 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_8_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_8_sparse(a)
    else:
        raise TypeError


def count_3motif_9(a):
    """Counts abundance of 3-nodes motif 9.
    :param a np.ndarray: adjacency matrix
    :return: motif 9 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_9_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_9_sparse(a)
    else:
        raise TypeError


def count_3motif_10(a):
    """Counts abundance of 3-nodes motif 10.
    :param a np.ndarray: adjacency matrix
    :return: motif 10 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_10_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_10_sparse(a)
    else:
        raise TypeError


def count_3motif_11(a):
    """Counts abundance of 3-nodes motif 11.
    :param a np.ndarray: adjacency matrix
    :return: motif 11 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_11_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_11_sparse(a)
    else:
        raise TypeError


def count_3motif_12(a):
    """Counts abundance of 3-nodes motif_ 12.
    :param a np.ndarray: adjacency matrix
    :return: motif 12 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_12_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_12_sparse(a)
    else:
        raise TypeError


def count_3motif_13(a):
    """Counts abundance of 3-nodes motif_ 13.
    :param a np.ndarray: adjacency matrix
    :return: motif 13 count
    :rtype: int
    """
    if isinstance(a, np.ndarray):
        return count_3motif_13_ndarray(a)
    elif isinstance(a, scipy.sparse.spmatrix):
        return count_3motif_13_sparse(a)
    else:
        raise TypeError



"""
# all tests works
"""
import sys

sys.path.append("../")
import NEMtropy.graph_classes as sample
from NEMtropy.models_functions import *
import NEMtropy.matrix_generator as mg
import numpy as np
import unittest  # test tool


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_iterative_0(self):

        n, seed = (4, 22)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)

        g = sample.DirectedGraph(a)
        g._solve_problem(
            model="dcm_exp",
            method="fixed-point",
            max_steps=3000,
            verbose=False,
            initial_guess="uniform",
            linsearch="False",
        )

        g._solution_error()
        err = g.error
        # debug
        # print('\ntest 1: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1)

    def test_iterative_1(self):

        n, seed = (4, 22)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a[0, :] = 0

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)

        g = sample.DirectedGraph(a)
        g._solve_problem(
            model="dcm_exp",
            method="fixed-point",
            max_steps=3000,
            verbose=False,
            initial_guess="uniform",
            linsearch="False",
        )

        g._solution_error()
        err = g.error
        # debug
        # print('\ntest 1: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1)

    @unittest.skip("skip large graph")
    def test_iterative_2(self):

        n, seed = (40, 22)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)

        g = sample.DirectedGraph(a)
        g._solve_problem(
            model="dcm_exp",
            method="fixed-point",
            max_steps=3000,
            verbose=False,
            initial_guess="uniform",
            linsearch="False",
        )

        g._solution_error()
        err = g.error
        # debug
        # print('\ntest 1: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1)

    @unittest.skip("skip large graph")
    def test_iterative_3(self):

        n, seed = (40, 22)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a[0, :] = 0

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)

        g = sample.DirectedGraph(a)
        g._solve_problem(
            model="dcm_exp",
            method="fixed-point",
            max_steps=3000,
            verbose=False,
            initial_guess="uniform",
            linsearch="False",
        )

        g._solution_error()
        err = g.error
        # debug
        # print('\ntest 1: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1)


if __name__ == "__main__":
    unittest.main()



import sys

sys.path.append("../")
import NEMtropy.graph_classes as sample
import NEMtropy.matrix_generator as mg
import numpy as np
import unittest  # test tool


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_0(self):
        n, seed = (4, 22)
        A = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        g = sample.DirectedGraph(A)

        g._solve_problem(
            model="dcm",
            method="newton",
            max_steps=300,
            verbose=False,
            initial_guess="uniform",
        )

        g._solution_error()
        # print('degseq = ', np.concatenate((g.dseq_out, g.dseq_in)))
        # print('expected degseq = ',g.expected_dseq)
        # print(np.concatenate((g.dseq_out, g.dseq_in)) - g.expected_dseq)
        # debug
        # print('\ntest 0  error = {}'.format(g.error))

        # test result
        self.assertTrue(g.error < 1)

    def test_1(self):
        n, seed = (4, 22)
        A = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        A[0, :] = 0

        g = sample.DirectedGraph(A)

        g._solve_problem(
            model="dcm",
            method="newton",
            max_steps=300,
            verbose=False,
            initial_guess="uniform",
        )

        g._solution_error()
        # print('degseq = ', np.concatenate((g.dseq_out, g.dseq_in)))
        # print('expected degseq = ',g.expected_dseq)
        # print(np.concatenate((g.dseq_out, g.dseq_in)) - g.expected_dseq)
        # debug
        # print('\ntest 1, n={}: error = {}'.format(n, g.error))

        # test result
        self.assertTrue(g.error < 1)

    @unittest.skip("skip large graph")
    def test_2(self):
        n, seed = (40, 22)
        A = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        g = sample.DirectedGraph(A)

        g._solve_problem(
            model="dcm",
            method="newton",
            max_steps=300,
            verbose=False,
            initial_guess="uniform",
        )

        g._solution_error()
        # print('degseq = ', np.concatenate((g.dseq_out, g.dseq_in)))
        # print('expected degseq = ',g.expected_dseq)
        # print(np.concatenate((g.dseq_out, g.dseq_in)) - g.expected_dseq)
        # debug
        print("\ntest 2, n={}: error = {}".format(n, g.error))

        # test result
        self.assertTrue(g.error < 1)

    @unittest.skip("skip large graph")
    def test_3(self):
        # test Matrix 1
        n, seed = (40, 22)
        A = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        g = sample.DirectedGraph(A)

        g._solve_problem(
            model="dcm",
            method="newton",
            max_steps=300,
            verbose=False,
            initial_guess="uniform",
        )

        g._solution_error()
        # print('degseq = ', np.concatenate((g.dseq_out, g.dseq_in)))
        # print('expected degseq = ',g.expected_dseq)
        # print(np.concatenate((g.dseq_out, g.dseq_in)) - g.expected_dseq)
        # debug
        print("\ntest 3, n={}: error = {}".format(n, g.error))

        # test result
        self.assertTrue(g.error < 1)

    @unittest.skip("skip large graph")
    def test_iterative_emi(self):

        n, seed = (50, 1)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        k_out = np.sum(a, 1)
        k_in = np.sum(a, 0)

        g = sample.DirectedGraph(a)
        g._solve_problem(
            model="dcm",
            method="newton",
            max_steps=3000,
            verbose=False,
            initial_guess="uniform",
            linsearch="False",
        )

        g._solution_error()
        err = g.error
        # debug
        print("\ntest emi: error = {}".format(err))

        # test result
        self.assertTrue(err < 1)


if __name__ == "__main__":
    unittest.main()
